%% Utrecht Summer School 2019 ABT Demo Congestion control model
%
% Demo illustrating how to branch off a Hopf point, branch off period % doubling bifurcations and continue Hopf bifurcation curve in two
% parameters. Furthermore, the stability of the various branches
% is computed.
%
%% Differential equations
%
% From: Hollot, C. V. and Chait, Y.
% Nonlinear stability analysis for a class of TCP/AQM networks
% Proceedings of the 40th IEEE Conference on Decision and Control, % 3:2309-2314, 2001
%
% $$\dot w(t)=1-\frac{w(t) w(t-1)}{2} k q(t-1),$$
%
% $$\dot q(t)=w(t)-c.$$
%
% Parameters are (in this order) |k|, |c|, |tau|
%% load DDE-BifTool into MATLAB path
clear
close all
ddebiftoolpath='../../'; addpath(strcat(ddebiftoolpath,'ddebiftool'),...
strcat(ddebiftoolpath,'ddebiftool_extra_psol'),... strcat(ddebiftoolpath,'ddebiftool_extra_nmfm'),...
strcat(ddebiftoolpath,'ddebiftool_utilities')); format compact
set(groot, 'defaultTextInterpreter', 'LaTeX');
%% Initial parameters and state
parnames={'k','c','tau'}; cind=[parnames;num2cell(1:length(parnames))];
ind=struct(cind{:});
bounds={'max_bound',[ind.k 8;ind.c 2.5],'max_step',[0,0.3],...
    'min_bound',[ind.k 2]};
%% Set user-defined functions
% use the right-hand side and derivatives created via symbolic toolbox 
funcs=set_symfuncs(@sym_congestionControlModel,'sys_tau',@()ind.tau);
%% Construct steady-state point
k=2; c=1; stst=dde_stst_create('x',[c;2/(k*c^2)]); stst.parameter(ind.k) = k; stst.parameter(ind.c) = c; stst.parameter(ind.tau) = 1;
% Compute stability
method_stst=df_mthod(funcs,'stst'); method_stst.stability.minimal_real_part=-3; stst.stability=p_stabil(funcs,stst,method_stst.stability);
% Plot eigenvalues
figure(1); clf;
plot(stst.stability.l1,'*')
title('Stability plot of stst')
xlabel('$\Re(\lambda)$')
ylabel('$\Im(\lambda)$')
%% Initialization of branch of steady-states
contpar=ind.k; steadystate_br=SetupStst(funcs,'x',stst.x,'parameter',stst.parameter,...
'step',0.1,'contpar',contpar,'max_step',[contpar,0.3],bounds{:});
%% Continue steady-state branch
figure(2);clf;ax2=gca;
n_steps=40; steadystate_br=br_contn(funcs,steadystate_br,n_steps,'plotaxis',ax2);
xlabel('$k$')
ylabel('$c$')
%% Detect bifurcations on steady-state branch
[steadystate_br,~,ind_hopf,bif1types]=LocateSpecialPoints(funcs,...
steadystate_br); nunst_eqs=GetStability(steadystate_br);
fprintf('Hopf bifurcation near point %d\n',ind_hopf);
%% Construct Hopf point manually
stst.parameter(ind.k) = 3.414105951101845;
stst.x = [1;0.585804901384075];
% Recompute stability
method_stst=df_mthod(funcs,'stst'); 
stst.stability=p_stabil(funcs,stst,method_stst.stability);
% Convert steady-state to Hopf point and compute the normal form 
hopf=p_tohopf(funcs,stst);
hopf=nmfm_hopf(funcs,hopf);
hopf.nmfm
 %% Branch off at Hopf bifurcation
disp('Branch off at Hopf bifurcation');
fprintf('Initial correction of periodic orbits at Hopf:\n'); [per1,suc]=SetupPsol(funcs,steadystate_br,ind_hopf,...
'print_residual_info',1,'intervals',20,'degree',4,...
'max_bound',[contpar,8],'max_step',[contpar,0.5],'matrix','full'); figure(3);clf;ax3=gca;
xlabel('$k$')
ylabel('Amplitude')
per1=br_contn(funcs,per1,60,'plotaxis',ax3);
%% Manually construct periodic solution near Hopf point
radius=0.1;
degree=4;
int_nr=20;
[psol,stpcond]=p_topsol(funcs,hopf,radius,degree,int_nr); method=df_mthod('psol'); [psol,succ]=p_correc(funcs,psol,ind.k,stpcond,method.point)
%% Compute stability of periodic solution
psol.stability=p_stabil(funcs,psol,method.stability);
%% Manual construction and continuation of psol branch 
per_orb=df_brnch(funcs,ind.k,'psol'); % empty branch: 
per_orb.parameter.max_bound=[ind.k 8];
per_orb.parameter.max_step=[ind.k 0.3]; 
deg_psol=p_topsol(funcs,hopf,0,degree,int_nr); 
per_orb.point=deg_psol;
per_orb.point(2)=psol;
% compute periodic solutions branch
per_orb=br_contn(funcs,per_orb,60,'plotaxis',ax3);
xlabel('$k$');
ylabel('amplitude');
%% Find period doubling bifurcation
per1=br_stabl(funcs,per1,0,1); 
nunst_per=GetStability(per1,'exclude_trivial',true); 
ind_pd=find(diff(nunst_per)==1);
%% Branch switching at period doubling bifurcation
ind_pd=find(diff(nunst_per)==1); 
[per2,suc]=DoublePsol(funcs,per1,ind_pd(1)); 
per2=br_contn(funcs,per2,60,'plotaxis',ax3); 
nunst_per2=GetStability(per2,'funcs',funcs,'exclude_trivial',true);
%% Branch switching at period doubling bifurcation
ind_pd2=find(diff(nunst_per2)==1); 
[per3,suc]=DoublePsol(funcs,per2,ind_pd2(1)); 
per3=br_contn(funcs,per3,60,'plotaxis',ax3); 
nunst_per3=GetStability(per3,'funcs',funcs,'exclude_trivial',true);
%% Branch switching at period doubling bifurcation
ind_pd3=find(diff(nunst_per3)==1); 
[per4,suc]=DoublePsol(funcs,per3,ind_pd3(1)); 
per4=br_contn(funcs,per4,60,'plotaxis',ax3); 
nunst_per4=GetStability(per4,'funcs',funcs,'exclude_trivial',true);
%% Continue period doubling bifurcations in two parameters
[pdfuncs,pdbranch1,suc]=SetupPeriodDoubling(funcs,per1,ind_pd(1),...
    'contpar',[ind.k,ind.c],'dir',ind.k,'step',1e-1,bounds{:});
pdbranch1=br_contn(pdfuncs,pdbranch1,60,'plotaxis',ax2);
pdbranch1=br_rvers(pdbranch1); 
pdbranch1=br_contn(pdfuncs,pdbranch1,60,'plotaxis',ax2); 
nunst_pd1=GetStability(pdbranch1,'exclude_trivial',true,'funcs',pdfuncs);
xlabel('$k$'); ylabel('$c$')
%% Continue period doubling bifurcations in two parameters
[pdfuncs,pdbranch2,suc]=SetupPeriodDoubling(funcs,per2,ind_pd2(1),...
    'contpar',[ind.k,ind.c],'dir',ind.k,'step',1e-1,bounds{:});
pdbranch2=br_contn(pdfuncs,pdbranch2,60,'plotaxis',ax2);
pdbranch2=br_rvers(pdbranch2); 
pdbranch2=br_contn(pdfuncs,pdbranch2,60,'plotaxis',ax2); 
nunst_pd1=GetStability(pdbranch2,'exclude_trivial',true,'funcs',pdfuncs);
xlabel('$k$'); ylabel('$c$')
%% Continue period doubling bifurcations in two parameters
[pdfuncs,pdbranch3,suc]=SetupPeriodDoubling(funcs,per3,ind_pd3(1),...
    'contpar',[ind.k,ind.c],'dir',ind.k,'step',1e-1,bounds{:});
pdbranch3=br_contn(pdfuncs,pdbranch3,60,'plotaxis',ax2);
pdbranch3=br_rvers(pdbranch3); 
pdbranch3=br_contn(pdfuncs,pdbranch3,60,'plotaxis',ax2); 
nunst_pd1=GetStability(pdbranch3,'exclude_trivial',true,'funcs',pdfuncs);
xlabel('$k$'); ylabel('$c$')
%% Continue Hopf bifurcation in two parameters
[hbranch,suc]=SetupHopf(funcs,steadystate_br,ind_hopf,...
'contpar',[ind.k,ind.tau],'dir',ind.k,'step',1e-1,bounds{:});
hbranch=br_contn(funcs,hbranch,60,'plotaxis',ax2); 
hbranch=br_rvers(hbranch);
hbranch=br_contn(funcs,hbranch,60,'plotaxis',ax2);
%% Compute L1 coefficient
% to find if Hopf bifurcation is supercritical (L1<0) or subcritical (L1>0)
[hbranch,hopftests,hc2_indices,hc2_types]=... 
    LocateSpecialPoints(funcs,hbranch);
%% Two-parameter bifurcation diagram
% Assigning a name and color to a curve. Others are chosen automatically 
figure(4); clf;ax3=gca;hold(ax3,'on')
lg=Plot2dBranch(hbranch);
lg=Plot2dBranch(pdbranch1,'lgname','PD1',... 
    'funcs',pdfuncs,'oldlegend',lg);
lg=Plot2dBranch(pdbranch2,'lgname','PD2',... 
    'funcs',pdfuncs,'oldlegend',lg,'color',[0,0.5,0]);
lg=Plot2dBranch(pdbranch3,'lgname','PD3',... 
    'funcs',pdfuncs,'oldlegend',lg,'color',[0,0,0.5]);
title('Bifurcation diagram of congestion control model') 
xlabel('$k$')
ylabel('$c$');
grid on
%% Time profiles of the first component of the period doubling solutions
bifsols={pdbranch1.point,pdbranch2.point,pdbranch3.point};
figure(5); clf;
for k=1:3
    subplot(3,1,k);
    hold on
    for i=1:length(bifsols{k})
        plot(bifsols{k}(i).mesh*bifsols{k}(i).period,...
    bifsols{k}(i).profile(1,:),'-');
   end
hold off
box on
grid on
title(sprintf('PD%d: time profiles of period doubling',k));
xlabel('$t$');
   ylabel('$w$');
end